{Data Aware Numeric-String Control    }
{Version 1.1     Delphi 16    03-16-96}
{Roger Nyberg  roger@servtech.com     }

unit Dbnum;

interface

uses SysUtils, WinTypes, WinProcs, Messages, Classes, Controls, Forms,
  Graphics, Menus, StdCtrls, ExtCtrls, DB, DBTables, Mask ,MoneyCnv;

type

TDisplayType = (ftFixed,ftNumber,ftGeneral,ftCurrency);

type
{ TDBNum }
  TDBNum = class(TCustomMaskEdit)
  private
    FDataLink: TFieldDataLink;
    FCanvas: TControlCanvas;
    FAlignment: TAlignment;
    FFocused: Boolean;
    FTabOnEnterKey : Boolean;
    FTextMargin: Integer;
    FDecimals : byte;
    FDispCalendar : Boolean;
    FDisplayType : TdisplayType ;
    FDispType : TFloatFormat;
    FNegColor : Tcolor;
    OldColor :TColor;
    FValue : extended;
    procedure CalcTextMargin;
    procedure DataChange(Sender: TObject);
    procedure EditingChange(Sender: TObject);
    function GetDataField: string;
    function GetDataSource: TDataSource;
    function GetField: TField;
    function GetReadOnly: Boolean;
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    procedure SetDecimals(Value : byte);
    procedure SetDispCalendar(Value : Boolean);
    procedure SetDisplayType(Value : TdisplayType);
    procedure SetTabOnEnterKey(Value: Boolean);
    procedure SetFocused(Value: Boolean);
    procedure SetNegCol(Value:TColor);
    procedure SetReadOnly(Value: Boolean);
    procedure UpdateData(Sender: TObject);
    procedure SetValue(Value : extended);
    procedure WMLButtonDblClk(var Message: TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
    procedure WMCut(var Message: TMessage); message WM_CUT;
    procedure WMPaste(var Message: TMessage); message WM_PASTE;
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
  protected
    procedure Change; override;
    function  EditCanModify: Boolean; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
    procedure Reset; override;
    procedure FormatText;
   public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Field: TField read GetField;
    function AsMoneyStr:String;
  published
    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property CharCase;
    property Color;
    property Ctl3D;
    property DataField: string read GetDataField write SetDataField;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property Decimals : byte read FDecimals write SetDecimals;
    property DisplayCalendar :Boolean read FDispCalendar write SetDispCalendar;
    property DisplayType :TdisplayType read FDisplayType write SetDisplayType;
    property DragCursor;
    property DragMode;
    property EditMask;
    property Enabled;
    property Font;
    property MaxLength;
    property NegativeColor : Tcolor read FNegColor write SetNegCol;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PasswordChar;
    property PopupMenu;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly default False;
    property ShowHint;
    Property TabOnEnterKey:Boolean read FTabOnEnterKey write SetTabOnEnterKey ;
    property TabOrder;
    property TabStop;
    property Value : extended read FValue write SetValue;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick ;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

procedure Register;
var
Cname:TObject;

implementation

uses DBIErrs, DBITypes, Clipbrd, DBConsts, Dialogs,DateForm ;

{ TDBNum }

constructor TDBNum.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  inherited ReadOnly := True;
  FDataLink := TFieldDataLink.Create;
  FDataLink.Control := Self;
  FDataLink.OnDataChange := DataChange;
  FDataLink.OnEditingChange := EditingChange;
  FDataLink.OnUpdateData := UpdateData;
  CalcTextMargin;
  FValue := 0.0;
  OldColor:=Font.Color;
  case FDisplayType of
       ftNumber   : FdispType := ffNumber ;
       ftGeneral  : FdispType := ffGeneral ;
       ftFixed    : FdispType := ffFixed ;
       ftCurrency : FdispType := ffCurrency ;
  end;
 end;

destructor TDBNum.Destroy;
begin
  FDataLink.Free;
  FDataLink := nil;
  FCanvas.Free;
  inherited Destroy;
end;

procedure TDBNum.SetDispCalendar(Value : Boolean);
begin
  if FDispCalendar <> Value then
     FDispCalendar:=Value;
 end;

procedure TDBNum.SetNegCol (Value: Tcolor);
begin
  if FNegColor <> Value then
     FNegColor:=Value;
 end;

procedure TDBNum.SetValue(Value : extended);
begin
  if FValue <> Value then
     FValue:=Value;
 end;

function TDBNum.AsMoneyStr:String;
var
XText:string;
begin
try
if (FDataLink.Field.DataType in ([ftFloat,ftSmallInt,ftInteger,ftWord])) then
begin
XText:=Text;
result:=MoneyStr(XText);
end else result:='Not Numeric Data Type';
 except
 result:='ERROR'
 end;
end;

procedure TDBNum.SetDecimals (Value: byte);
begin
  if FDecimals <> Value then
     FDecimals := Value;
  if (MaxLength < FDecimals + 1) and (FDecimals<>0) then
     MaxLength:= FDecimals + 1;
     FormatText;
 end;

procedure TDBNum.FormatText;
var
TrashText:String;
i:Integer;
begin
    if (FDataLink.Field <> nil)
    and (FDataLink.Field.DataType in ([ftFloat,ftSmallInt,ftInteger,ftWord])) then
    begin
     TrashText:='';
    for i:=1 to length(Text) do
    if ( Text[i] in ['0'..'9',DecimalSeparator,'-']) then
         TrashText:=TrashText+Text[i];
    if TrashText='' then TrashText:='0';
    Text := FloatToStrF(StrToFloat(TrashText),FDispType, MaxLength, FDecimals);
   end;
end;

procedure TDBNum.SetTabOnEnterKey(Value:Boolean);
begin
  if FTabOnEnterKey <> Value then
  begin
    FTabOnEnterKey := Value;
  end;
end;

procedure TDBNum.SetDisplayType(Value: TdisplayType);
begin
  if FDisplayType <> Value then
   begin
       FDisplayType := Value;

      case FDisplayType of
      ftNumber   : FdispType := ffNumber ;
      ftGeneral  : FdispType := ffGeneral ;
      ftFixed    : FdispType := ffFixed ;
      ftCurrency : FdispType := ffCurrency ;
       end;
       FormatText;
    end
  end;

procedure TDBNum.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (FDataLink <> nil) and
    (AComponent = DataSource) then DataSource := nil;
end;

procedure TDBNum. WMLButtonDblClk(var Message: TWMLButtonDblClk);
 begin
 if (FDataLink.Field <> nil)
 and (FDataLink.Field.DataType = ftDate)
 and (FDispCalendar) then
  begin
   inherited DblClick;
    FDataLink.Edit;
    Cname:=Self;
    Application.CreateForm(TDtForm, DtForm);
    DtForm.Showmodal;
    DtForm.Free;
  end else
  inherited
 end;

procedure TDBNum.KeyDown(var Key: Word; Shift: TShiftState);
begin
   if (key = VK_UP) then
      PostMessage(GetparentForm(Self).Handle, WM_NEXTDLGCTL ,1,0);
   if (Key = VK_DOWN) then
      PostMessage(GetparentForm(Self).Handle, WM_NEXTDLGCTL ,0,0);
  inherited KeyDown(Key, Shift);
  if (Key = VK_DELETE) or ((Key = VK_INSERT) and (ssShift in Shift)) then
    FDataLink.Edit;
end;

procedure TDBNum.KeyPress(var Key: Char);
var
TrashText,OldText:String;
i,x,Sel,OldLength:Integer;
begin

{if EnterKey and TabOnEnter}
  if (key = #13)and (FTabOnEnterKey)then
   begin
       PostMessage(GetparentForm(Self).Handle, WM_NEXTDLGCTL ,0,0);
       Key := #0;
       Exit;
   end;

 inherited KeyPress(Key);

 if (Key in [#32..#255]) and (FDataLink.Field <> nil) and
    not FDataLink.Field.IsValidChar(Key) then
  begin
    MessageBeep(0);
    Key := #0;
  end;
  case Key of
    ^H, ^V, ^X, #32..#255:
      FDataLink.Edit;
    #27:
      begin
        FDataLink.Reset;
        SelectAll;
        Key := #0;
      end;
    end;

 {Run this if field is float or integer type}
if  (FDataLink.Field.DataType in ([ftFloat,ftSmallInt,ftInteger,ftWord]))
    and (Key <> #0) then
    begin
{if copy or cut selection}
if (Key = ^C) or (Key = ^X) then exit;

{if Paste check if valid else undo}
if (Key = ^V) then
begin
 try
 OldText:=Text;
 SendMessage(Self.Handle, WM_PASTE ,0,0);
 FormatText;
  if Length(Text) > MaxLength then
   begin
    Text:=OldText;
    MessageBeep(0);
   end;
  except
    SendMessage(Self.Handle, WM_UNDO ,0,0);
    MessageBeep(0);
  end;
 Key:=#0;
 exit;
end;

{Check for allowed characters}
 if not (key in ['0'..'9',DecimalSeparator,'-',#8]) then
      begin
        MessageBeep(0);
        Key:=#0;
        exit;
      end;

 {Clear selection in control If selection include DecimalPoint clear
   all numbers from start selection}
  if Sellength > 0 then
  begin
    if (SelStart+1 <= pos(DecimalSeparator,Text))
    and (SelStart+SelLength >= pos(DecimalSeparator,Text))then
    begin
    SelLength:=Length(Text)- SelStart;
    SendMessage(Self.Handle, WM_CLEAR, 0, 0);
    end
    else
    SendMessage(Self.Handle, WM_CLEAR, 0, 0);
   end;

 {Take away all non numeric characters and leave the cursor}
   x:=0;
   Sel:=SelStart;
   TrashText:='';
   OldLength:=Length(Text);
    for i:=1 to length(Text) do
    if ( Text[i] in ['0'..'9',DecimalSeparator,'-']) then
       TrashText:=TrashText+Text[i] else if i > sel then inc(x);
       Text:=TrashText;
       SelStart:=Sel-(OldLength-Length(Text))+x;

 {Check for Back Space on Decimal, if decimal exist restrict removal if it will
 exceed characters before Decimal}
if     (key=#8)
       and (pos(DecimalSeparator,Text) >0)
       and (SelStart = pos(DecimalSeparator,Text))
       and (Length(Text)> pos(DecimalSeparator,Text))
       and (Length(Text)> MaxLength-FDecimals) then
       begin
        MessageBeep(0);
        Key:=#0;
        Exit;
       end;

 {Check for decimal is allowed and if decimal exist}
 if   (key = DecimalSeparator)
      and (Fdecimals = 0)
      or  (pos(DecimalSeparator,Text) >0)
      and (Key=DecimalSeparator)
      or  (SelStart<Length(Text)-FDecimals)
      and (Key=DecimalSeparator)
      and (Length(Text)>FDecimals)then
      begin
        MessageBeep(0);
        Key:=#0;
        Exit;
      end;

 {Check for negative sign, only allowed as first character}
  if (key='-')
      and (SelStart<>0)
      or (key='-')
      and (pos('-',Text) >0)then
       begin
        MessageBeep(0);
        Key:=#0;
        Exit;
       end;

 {Check max characters before decimal }
        if  (Key<>#8)
             and (pos(DecimalSeparator,Text) >0)
             and (pos(DecimalSeparator,Text) >= MaxLength-FDecimals)
             and (selstart < pos(DecimalSeparator,Text))
             or  (Key in ['0'..'9'])
             and (FDecimals <>0)
             and (selstart <= MaxLength-FDecimals)
             and (Length(Text)>=MaxLength-FDecimals-1)
             and (pos(DecimalSeparator,Text) = 0)
             then
                begin
      {Add Decimal automatically if max characters before Decimal}
                if (Key in ['0'..'9'])
                 and (FDecimals <>0)
                 and (selstart = MaxLength-FDecimals-1)
                 and (pos(DecimalSeparator,Text) = 0) then
                    begin
                         key:=DecimalSeparator;
                         exit
                    end;
                   MessageBeep(0);
                   Key:=#0;
                   Exit;
                end;

  {Check max characters after decimal}
        if  (Key<>#8)
             and (pos(DecimalSeparator,Text) >0)
             and (Length(Text)-pos(DecimalSeparator,Text)>=FDecimals)
             and (selstart >= pos(DecimalSeparator,Text)) then
                begin
                    MessageBeep(0);
                     Key:=#0;
                     Exit;
                end;

     end;
end;

function TDBNum.EditCanModify: Boolean;
begin
  Result := FDataLink.Edit;
end;

procedure TDBNum.Reset;
begin
  FDataLink.Reset;
  SelectAll;
end;

procedure TDBNum.SetFocused(Value: Boolean);
begin
  if FFocused <> Value then
  begin
    FFocused := Value;
    if (FAlignment <> taLeftJustify) and not IsMasked then Invalidate;
    FDataLink.Reset;
  end;
end;

procedure TDBNum.Change;
begin
  FDataLink.Modified;
  inherited Change;
end;

function TDBNum.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

procedure TDBNum.SetDataSource(Value: TDataSource);
begin
  FDataLink.DataSource := Value;
end;

function TDBNum.GetDataField: string;
begin
  Result := FDataLink.FieldName;
end;

procedure TDBNum.SetDataField(const Value: string);
begin
  FDataLink.FieldName := Value;
end;

function TDBNum.GetReadOnly: Boolean;
begin
  Result := FDataLink.ReadOnly;
end;

procedure TDBNum.SetReadOnly(Value: Boolean);
begin
  FDataLink.ReadOnly := Value;
end;

function TDBNum.GetField: TField;
begin
  Result := FDataLink.Field;
end;

procedure TDBNum.DataChange(Sender: TObject);
var
TrashText:String;
begin
  if FDataLink.Field <> nil then
  begin
    if FAlignment <> FDataLink.Field.Alignment then
    begin
      EditText := '';  {forces update}
      FAlignment := FDataLink.Field.Alignment;
    end;

{If Field Mask exist use it}
  if (FDataLink.Field.EditMask <>'') then
      EditMask := FDataLink.Field.EditMask;

{If field is a numeric field clear the mask}
if (FDataLink.Field.DataType in [ftFloat,ftSmallInt,ftInteger,ftWord]) then
 EditMask:='';

{Set MaxLength for String field}
if FDataLink.Field.DataType = ftString then
   MaxLength := FDataLink.Field.Size;

{If integer type set Decimals to 0}
 if  FDataLink.Field.DataType in ([ftSmallInt,ftInteger,ftWord]) then
     FDecimals:=0;
    if FFocused and FDataLink.CanModify then
      if  FDataLink.Field.DataType in ([ftFloat,ftSmallInt,ftInteger,ftWord]) then
                  if FDataLink.Field.Text = '' then
                     begin
                     Text := FloatToStrF(0,FDispType, MaxLength, FDecimals);
                     FValue := 0.0;
                end else
                  begin
                   TrashText:=FDataLink.Field.Text;
                   while pos(ThousandSeparator,trashText)>0 do
                         delete(TrashText,pos(ThousandSeparator,TrashText),1);
      Text := FloatToStrF(StrToFloat(TrashText),FDispType, MaxLength, FDecimals);
      FValue := StrToFloat(TrashText);
      if FValue < 0 then Font.Color:=FNegColor else Font.Color:=OldColor;
                  end
              else
           Text := FDataLink.Field.Text
      else
           if  FDataLink.Field.DataType in ([ftFloat,ftSmallInt,ftInteger,ftWord]) then
                    if FDataLink.Field.DisplayText = '' then
                     begin
                       EditText := FloatToStrF(0,FDispType, MaxLength, FDecimals);
                        FValue := 0.0;
                   end else
                     begin
                     TrashText:=FDataLink.Field.DisplayText;
                     while pos(ThousandSeparator,TrashText)>0 do
                           delete(TrashText,pos(ThousandSeparator,TrashText),1);
      EditText := FloatToStrF(StrToFloat(TrashText),FDispType, MaxLength, FDecimals);
      FValue := StrToFloat(TrashText);
      if FValue < 0 then Font.Color:=FNegColor else Font.Color:=OldColor;
                     end
                  else
            EditText := FDataLink.Field.DisplayText;
  end else
  begin
    FAlignment := taLeftJustify;
    if csDesigning in ComponentState then
      EditText := Name else
      EditText := '';
  end;
end;

procedure TDBNum.EditingChange(Sender: TObject);
begin
  inherited ReadOnly := not FDataLink.Editing;
end;

procedure TDBNum.UpdateData(Sender: TObject);
var
TrashText : String;
i:integer;
begin
 ValidateEdit;
 TrashText := Text;

 {Remove all non numeric characters before saving}
if  FDataLink.Field.DataType in ([ftFloat,ftSmallInt,ftInteger,ftWord]) then
 begin
     TrashText:='';
    for i:=1 to length(Text) do
    if ( Text[i] in ['0'..'9',DecimalSeparator,'-']) then
         TrashText:=TrashText+Text[i];
  if (TrashText='') or (TrashText='-')  then TrashText:='0';
 end;
  FDataLink.Field.Text := TrashText;
end;

procedure TDBNum.WMPaste(var Message: TMessage);
begin
  FDataLink.Edit;
  inherited;
end;

procedure TDBNum.WMCut(var Message: TMessage);
begin
  FDataLink.Edit;
  inherited;
end;

procedure TDBNum.CMEnter(var Message: TCMEnter);
begin
  SetFocused(True);
  inherited;
end;

procedure TDBNum.CMExit(var Message: TCMExit);
begin
  try
 {Check for empty Date Field, if empty clear the field}
    if (FDataLink.Field.Datatype =ftDate)
    and (Text = '  /  /  ')
    and (FDataLink.Editing)
    or (FDataLink.Field.Datatype =ftDate)
    and (Text = '')
    and (FDataLink.Editing) then
    FDataLink.Field.Clear;

    FDataLink.UpdateRecord;
  except
    SelectAll;
    SetFocus;
    raise;
  end;
  SetFocused(False);
  CheckCursor;
  DoExit;
end;

procedure TDBNum.WMPaint(var Message: TWMPaint);
var
  Width, Indent, Left, I: Integer;
  R: TRect;
  DC: HDC;
  PS: TPaintStruct;
  S: string;
begin
  if (FAlignment = taLeftJustify) or FFocused then
  begin
    inherited;
    Exit;
  end;
{ Since edit controls do not handle justification unless multi-line (and
  then only poorly) we will draw right and center justify manually unless
  the edit has the focus. }
  if FCanvas = nil then
  begin
    FCanvas := TControlCanvas.Create;
    FCanvas.Control := Self;
  end;
  DC := Message.DC;
  if DC = 0 then DC := BeginPaint(Handle, PS);
  FCanvas.Handle := DC;
  try
    FCanvas.Font := Font;
    with FCanvas do
    begin
      R := ClientRect;
      if (BorderStyle = bsSingle) then
      begin
        Brush.Color := clWindowFrame;
        FrameRect(R);
        InflateRect(R, -1, -1);
      end;
      Brush.Color := Color;
      S := EditText;
      if PasswordChar <> #0 then
      begin
        for I := 1 to Length(S) do
          S[I] := PasswordChar;
      end;
      Width := TextWidth(S);
      if BorderStyle = bsNone then Indent := 0 else Indent := FTextMargin;
      if FAlignment = taRightJustify then
        Left := R.Right - Width - Indent else
        Left := (R.Left + R.Right - Width) div 2;
      TextRect(R, Left, Indent, S);
    end;
  finally
    FCanvas.Handle := 0;
    if Message.DC = 0 then EndPaint(Handle, PS);
  end;
end;

procedure TDBNum.CMFontChanged(var Message: TMessage);
begin
  inherited;
  CalcTextMargin;
  end;

procedure TDBNum.CalcTextMargin;
var
  DC: HDC;
  SaveFont: HFont;
  I: Integer;
  SysMetrics, Metrics: TTextMetric;
begin
  DC := GetDC(0);
  GetTextMetrics(DC, SysMetrics);
  SaveFont := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, SaveFont);
  ReleaseDC(0, DC);
  I := SysMetrics.tmHeight;
  if I > Metrics.tmHeight then I := Metrics.tmHeight;
  FTextMargin := I div 4;
end;

procedure Register;
begin
  RegisterComponents ('Data Controls', [TDBNum]);
end;
end.

